/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import com.mycila.xmltool.CallBack;
import com.mycila.xmltool.ValidationResult;
import com.mycila.xmltool.XMLDoc;
import com.mycila.xmltool.XMLTag;
import com.ovea.resource.web.util.Resource;
import com.ovea.resource.web.util.Scanner;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;

import javax.servlet.ServletContext;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.*;

final class XmlBasedWebResourceProvider {

    private final ServletContext servletContext;
    private final String descriptor;
    private volatile VirtualResources virtualResources;
    private boolean debug;
    private Properties properties = new Properties();

    XmlBasedWebResourceProvider(ServletContext servletContext, String descriptor) {
        this.servletContext = servletContext;
        this.descriptor = descriptor;
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    public VirtualResources get() {
        if (virtualResources == null || debug) {
            VirtualResources virtualResources = new VirtualResources();
            virtualResources.setDebug(debug);
            virtualResources.setProperties(properties);
            Enumeration<URL> resources;
            try {
                resources = Thread.currentThread().getContextClassLoader().getResources(descriptor);
            } catch (IOException e) {
                throw new RuntimeException(e.getMessage(), e);
            }
            while (resources.hasMoreElements()) {
                append(virtualResources, resources.nextElement());
            }
            this.virtualResources = virtualResources;
        }
        return virtualResources;
    }

    private void append(final VirtualResources virtualResources, URL url) {
        ValidationResult result = XMLDoc.from(url, false).validate(Resource.from("classpath:com/ovea/resource/web/virtual-resources.xsd").toURL());
        if (result.hasError())
            throw new IllegalStateException("Document " + url + " is not valid: " + Arrays.deepToString(result.getErrorMessages()), result.getErrors()[0]);
        XMLTag tag = XMLDoc.from(url, true);
        // <map>
        tag.forEach("map", new CallBack() {
            @Override
            public void execute(final XMLTag mapTag) {
                String b = mapTag.getAttribute("virtual-path");
                if (b.startsWith("/"))
                    b = b.substring(1);
                if (b.endsWith("/"))
                    b = b.substring(0, b.length() - 1);
                final String newBasedir = b;
                final boolean optional = mapTag.hasAttribute("optional") ? Boolean.valueOf(mapTag.getAttribute("optional")) : false;
                String to = mapTag.getAttribute("to");
                if (to.startsWith("web:")) {
                    to = "file:" + new File(servletContext.getRealPath(""), to.substring(4)).getAbsolutePath();
                }
                Scanner.scan(Resource.from(to), new Scanner.Callback() {
                    @Override
                    public void onMatch(String relativePath, Resource resource) {
                        VirtualResource virtualResource = virtualResources.addVirtualResource(newBasedir + relativePath);
                        WebResource wr = virtualResource.addWebResource(resource);
                        wr.setOptional(optional);
                        if (mapTag.hasAttribute("caching"))
                            virtualResource.setCaching(Long.parseLong(mapTag.getAttribute("caching")));
                        if (mapTag.hasAttribute("gzip"))
                            virtualResource.setGzip(Boolean.parseBoolean(mapTag.getAttribute("gzip")));
                    }

                    @Override
                    public void onError(Throwable t) {
                        if (!optional) {
                            if (t instanceof RuntimeException)
                                throw (RuntimeException) t;
                            if (t instanceof Error)
                                throw (Error) t;
                            throw new RuntimeException(t.getMessage(), t);
                        }
                    }
                });
            }
        });
        // <resource>
        tag.forEach("resource", new CallBack() {
            public void execute(XMLTag resTag) {
                final VirtualResource virtualResource = virtualResources.addVirtualResource(resTag.getAttribute("virtual-path"));
                if (resTag.hasAttribute("contentType"))
                    virtualResource.setContentType(resTag.getAttribute("contentType"));
                if (resTag.hasAttribute("caching"))
                    virtualResource.setCaching(Long.parseLong(resTag.getAttribute("caching")));
                if (resTag.hasAttribute("gzip"))
                    virtualResource.setGzip(Boolean.parseBoolean(resTag.getAttribute("gzip")));
                resTag.forEachChild(new CallBack() {
                    public void execute(XMLTag locTag) {
                        boolean optional = locTag.hasAttribute("optional") && Boolean.parseBoolean(locTag.getAttribute("optional"));
                        List<Resource> resources = buidResources(virtualResource.getVirtualPath(), locTag);
                        if(resources.isEmpty() && !optional) {
                            throw new IllegalArgumentException("No resource found matching " + locTag.getText() + " in virtual path " + virtualResource.getVirtualPath());
                        }
                        for (Resource resource : resources) {
                            WebResource webResource = virtualResource.addWebResource(resource);
                            webResource.setOptional(optional);
                            if (locTag.hasAttribute("minimize"))
                                webResource.setMinimize(Boolean.parseBoolean(locTag.getAttribute("minimize")));
                        }
                    }

                    private List<Resource> buidResources(String vPath, XMLTag res) {
                        switch (WebResourceLocation.from(res.getCurrentTagName())) {
                            case CLASSPATH: {
                                String path = res.getText();
                                if (path.startsWith("/"))
                                    path = path.substring(1);
                                PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(Thread.currentThread().getContextClassLoader());
                                try {
                                    List<Resource> resourceList = new LinkedList<Resource>();
                                    for (org.springframework.core.io.Resource resource : resolver.getResources("classpath*:" + path)) {
                                        resourceList.add(Resource.from(resource.getURL()));
                                    }
                                    return resourceList;
                                } catch (IOException e) {
                                    throw new RuntimeException("Error loading resources for virtual path " + vPath + " : " + e.getMessage(), e);
                                }
                            }
                            case FILE: {
                                String path = res.getText();
                                PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(Thread.currentThread().getContextClassLoader());
                                try {
                                    List<Resource> resourceList = new LinkedList<Resource>();
                                    for (org.springframework.core.io.Resource resource : resolver.getResources("file:" + path)) {
                                        resourceList.add(Resource.from(resource.getURL()));
                                    }
                                    return resourceList;
                                } catch (IOException e) {
                                    throw new RuntimeException("Error loading resources for virtual path " + vPath + " : " + e.getMessage(), e);
                                }
                            }
                            case WEB: {
                                String webPath = servletContext.getRealPath("");
                                String path = res.getText();
                                if (path.startsWith("/"))
                                    path = path.substring(1);
                                PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(Thread.currentThread().getContextClassLoader());
                                try {
                                    List<Resource> resourceList = new LinkedList<Resource>();
                                    for (org.springframework.core.io.Resource resource : resolver.getResources("file:" + webPath + "/" + path)) {
                                        resourceList.add(Resource.from(resource.getURL()));
                                    }
                                    return resourceList;
                                } catch (IOException e) {
                                    throw new RuntimeException("Error loading resources for virtual path " + vPath + " : " + e.getMessage(), e);
                                }
                            }
                            case CODE: {
                                return Arrays.asList(Resource.fromData(res.getCDATAorText()));
                            }
                        }
                        throw new IllegalArgumentException(res.getCurrentTagName());
                    }
                });
            }
        });
    }

}